/*!
 * @file    uwb_SmartField_IO.h
 * @author  Kryštof Vaněk <krystof.vanek@gmail.com>
 * @version 1.0
 * @date    2021-02-15
 * @brief   SmartField IO driver prototype
 *
 * @attention <h2><center>&copy; Copyright (c) 2021 University of West Bohemia in Pilsen All rights reserved.</center></h2>
 * @attention <h2><center>Use of this source code without author permission and license agreement is prohibited. For license agreement contact the author.</center></h2>
 * @attention <h2><center>Tento zdrojový kód nelze využívat bez souhlasu autora a dohodě o licenčních podmínkách. Pro případné licencování kontaktujte autora.</center></h2>
 *
 * Developed by:<br>
 * University of West Bohemia<br>
 * Department of Technologies and Measurement<br>
 * www.zcu.cz
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 * OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
 */

#ifndef _UWB_SMARTFIELD_IO_H_
#define _UWB_SMARTFIELD_IO_H_

#ifdef __cplusplus
extern "C" {
#endif

typedef enum _SmartField_IO_Status{
	SFIO_OK=0,
	SFIO_NOT_IMPLEMENTED,		/*!< Returned by weak functions, module has to evaluate if it is considered an error */
	SFIO_ERROR,					/*!< General error, all errors have greater value (used for simple output check status<error) */
	SFIO_TARGET_UNDEFINED,
	SFIO_HAL_ERROR
} SmartField_IO_Status_t;

typedef struct _SmartField_IO_Time {
	uint8_t bHours;
	uint8_t bMinutes;
	uint8_t bSeconds;
	uint8_t bDay;
	uint8_t bMonth;
	uint8_t bYear;
} SmartField_IO_Time_t;

typedef struct _SmartField_IO_terminalColors{
	const char *reset;
	const char *red;
	const char *green;
	const char *blue;
	const char *cyan;
	const char *magenta;
	const char *yellow;
	const char *black;
	const char *white;
} SmartField_IO_TerminalColors_t;

extern SmartField_IO_TerminalColors_t terminalColors;

#define SmartField_printf(...) \
	SmartField_IO_TerminalTimeStamp(); \
	printf(" > "); \
	printf(__VA_ARGS__)

#define SmartField_printf_FAT(...) \
	SmartField_IO_TerminalTimeStamp(); \
	printf("  [FAT] > "); \
	printf(__VA_ARGS__)

#define SmartField_printf_API(...) \
	SmartField_IO_TerminalTimeStamp(); \
	printf("  [API] > "); \
	printf(__VA_ARGS__)

#define SmartField_printf_Core(...) \
	SmartField_IO_TerminalTimeStamp(); \
	printf(" [Core] > "); \
	printf(__VA_ARGS__)

#define SmartField_printf_PLL(...) \
	SmartField_IO_TerminalTimeStamp(); \
	printf("  [PLL] > "); \
	printf(__VA_ARGS__)

#define SmartField_printf_Ident(...) \
	SmartField_IO_TerminalTimeStamp(); \
	printf("\033[0;31m[Ident]\033[0m > "); \
	printf(__VA_ARGS__)

/**
 * @defgroup SmartField_IO SmartField IO module prototype
 * @brief SmartField Input-Output module declaration and weak definition
 * @{
 */

/**
 * @defgroup SmartField_IO_LEDs SmartField LED management
 * @brief SmartField LED management module declaration and weak definition
 *
 * These functions are optional and doesn't have to be implemented.
 *
 * @{
 */
extern SmartField_IO_Status_t SmartField_IO_LED_STS_ON();
extern SmartField_IO_Status_t SmartField_IO_LED_STS_OFF();
extern SmartField_IO_Status_t SmartField_IO_LED_PRB_ON();
extern SmartField_IO_Status_t SmartField_IO_LED_PRB_OFF();
extern SmartField_IO_Status_t SmartField_IO_LED_ERR_ON();
extern SmartField_IO_Status_t SmartField_IO_LED_ERR_OFF();
/**
 * @}
 */

/**
 * @defgroup SmartField_IO_PWR SmartField power management
 * \ingroup SmartField_IO
 * @brief SmartField power management module declaration and weak definition
 * @attention User must provide definitions of declared functions in application!
 * @{
 */
extern SmartField_IO_Status_t SmartField_IO_VDD_Ident_ON();
extern SmartField_IO_Status_t SmartField_IO_VDD_Ident_OFF();
extern SmartField_IO_Status_t SmartField_IO_VCC_ON();
extern SmartField_IO_Status_t SmartField_IO_VCC_OFF();
extern SmartField_IO_Status_t SmartField_IO_VDD_PLL_ON();
extern SmartField_IO_Status_t SmartField_IO_VDD_PLL_OFF();
/**
 * @}
 */

/**
 * @defgroup SmartField_IO_ACT SmartField subsystem activation management
 * \ingroup SmartField_IO
 * @brief SmartField subsystem activation management module declaration and weak definition
 * @attention User must provide definitions of declared functions in application!
 * @{
 */
extern SmartField_IO_Status_t SmartField_IO_GEN_Enable();
extern SmartField_IO_Status_t SmartField_IO_GEN_Disable();
extern SmartField_IO_Status_t SmartField_IO_CH1_Enable();
extern SmartField_IO_Status_t SmartField_IO_CH1_Disable();
extern SmartField_IO_Status_t SmartField_IO_CH2_Enable();
extern SmartField_IO_Status_t SmartField_IO_CH2_Disable();
extern SmartField_IO_Status_t SmartField_IO_I2C_Enable();
extern SmartField_IO_Status_t SmartField_IO_I2C_Disable();
/**
 * @}
 */


extern SmartField_IO_Status_t SmartField_IO_SampleData(uint16_t* pBuff, uint32_t dataLength);

/**
 * @defgroup SmartField_IO_Storage SmartField file storage
 * \ingroup SmartField_IO
 * @brief SmartField file storage module declaration and weak definition
 *
 * These functions are optional and doesn't have to be implemented.
 * @{
 */
extern SmartField_IO_Status_t SmartField_IO_SDcard_writeHeader(SmartField_Settings_t* settings);
extern SmartField_IO_Status_t SmartField_IO_SDcard_writeData(uint8_t *pBuff, uint32_t dataLength);
extern SmartField_IO_Status_t SmartField_IO_SDcard_finalise();
/**
* @}
*/

/**
* @}
*/

extern void SmartField_IO_TerminalTimeStamp();

#ifdef __cplusplus
}
#endif

#endif /* _UWB_SMARTFIELD_IO_H_ */
