/*!
 * @file    uwb_SmartField_core.c
 * @author  Kryštof Vaněk <krystof.vanek@gmail.com>
 * @version 3.0
 * @date    2021-02-15
 * @brief   SmartField computation core
 *
 * @attention <h2><center>&copy; Copyright (c) 2021 University of West Bohemia in Pilsen All rights reserved.</center></h2>
 * @attention <h2><center>Use of this source code without author permission and license agreement is prohibited. For license agreement contact the author.</center></h2>
 * @attention <h2><center>Tento zdrojový kód nelze využívat bez souhlasu autora a dohodě o licenčních podmínkách. Pro případné licencování kontaktujte autora.</center></h2>
 *
 * Developed by:<br>
 * University of West Bohemia<br>
 * Department of Technologies and Measurement<br>
 * www.zcu.cz
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 * OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
 */

/* Includes ------------------------------------------------------------------*/
#include "uwb_SmartField.h"
#include "uwb_SmartField_firCoef.h"
#include <tgmath.h>

/* Private typedef -----------------------------------------------------------*/
typedef struct {
    float 	interpPos;
    bool 	rising;
} SmartField_Signal_Zero_t;

/* Private define ------------------------------------------------------------*/

/* Private macro -------------------------------------------------------------*/

/* Private variables ---------------------------------------------------------*/
static uint8_t 				SmartField_ADC_RAW_Buffer_Byte[SMARTFIELD_ADC_BUFFER_LENGTH][2]={0};
static SmartField_Sample_t* SmartField_ADC_RAW_Buffer = (SmartField_Sample_t*)&SmartField_ADC_RAW_Buffer_Byte[0][0];
static float  				SmartField_Voltage[SMARTFIELD_FILT_BUFFER_LENGTH]={0};
static float  				SmartField_Current[SMARTFIELD_FILT_BUFFER_LENGTH]={0};
static SmartField_Result_t 	SmartField_Result={0};

/* Private function prototypes -----------------------------------------------*/
static SmartField_Status_t	checkData(void);
static SmartField_Status_t	checkVector(uint8_t pArray[][2],  uint8_t colIndex);
static void					filterData(uint8_t pData[][2], uint8_t colIndex, float *pResult);
static float 				timeDomainPhase(SmartField_Signal_Zero_t *x, SmartField_Signal_Zero_t*y, uint32_t zerosFound);
static uint32_t				findZeros(float *vector, SmartField_Signal_Zero_t *zeros, uint32_t maxZeros);
static float _Complex 		goertzel(uint8_t pData[][2], uint8_t colIndex, uint32_t k);

/* Public function definitions -----------------------------------------------*/

/**
 * @brief Load data samples to static buffer
 *
 * @retval	SMARTFIELD_OK	In case of success.
 * @retval  SMARTFIELD_ERR	In case of IO layer error.
 *
 * Function calls function SmartField_IO_SampleData() with private buffer pointer and requested number of samples.
 */
SmartField_Status_t SmartField_Core_SampleData(void) {

	if(SFIO_OK != SmartField_IO_SampleData((uint16_t*)SmartField_ADC_RAW_Buffer, SMARTFIELD_ADC_BUFFER_LENGTH)){
		return SMARTFIELD_ERR;
	}

	return SMARTFIELD_OK;
}

/**
 * @brief Get raw data from buffer
 *
 * @param	callbackPtr	Function pointer to be called.
 *
 * Function calls callback given in pointer, providing data buffer location and length.
 */
void SmartField_Core_DumpRawData(void (*callbackPtr)(uint8_t* pData, uint32_t dataLength)) {
	if(callbackPtr!=NULL){
		callbackPtr((uint8_t*)SmartField_ADC_RAW_Buffer_Byte, 2*SMARTFIELD_ADC_BUFFER_LENGTH);
	}
}

/**
 * @brief Calculate impedance using FIR filter and time-domain phase evaluation
 *
 * @param	settings	SmartField configuration pointer
 *
 * Checks data using checkData() and applies FIR filtering using filterData().
 * Phase delay between voltage and current vectors is measured by timeDomainPhase().
 * Complex impedance is calculated from voltage and current RMS and phase.
 */
void SmartField_Core_ProcessData(SmartField_Settings_t* settings){
	checkData();

//	#ifdef DEBUG
//		SmartField_printf_Core("Filtering voltage data...");
//	#endif
	filterData(SmartField_ADC_RAW_Buffer_Byte, SMARTFIELD_VOLTAGE_COL, SmartField_Voltage);
//	#ifdef DEBUG
//		SmartField_printf_Core("Filtering current data...");
//	#endif
	filterData(SmartField_ADC_RAW_Buffer_Byte, SMARTFIELD_CURRENT_COL, SmartField_Current);

	//Calibrate voltage data
	for (uint16_t i = 0; i < SMARTFIELD_FILT_BUFFER_LENGTH; i++) {
		SmartField_Voltage[i] = SmartField_Voltage[i] * SMARTFIELD_Q;
	}

	//Calibrate current data
	for (uint16_t i = 0; i < SMARTFIELD_FILT_BUFFER_LENGTH; i++) {
		SmartField_Current[i] = -SmartField_Current[i] * SMARTFIELD_Q/settings->Zsns;
	}

	//Calculate RMS values
	SmartField_Result.VRMS=0;
	SmartField_Result.IRMS=0;
	for (uint16_t i = 0; i < 1024; i++) {
		SmartField_Result.VRMS += pow(SmartField_Voltage[i], (float) 2);
		SmartField_Result.IRMS += pow(SmartField_Current[i], (float) 2);
	}
	SmartField_Result.VRMS = sqrt(SmartField_Result.VRMS / 1024);
	SmartField_Result.IRMS = sqrt(SmartField_Result.IRMS / 1024);

    //Calculate angle
    uint32_t maxZeros=10;
    SmartField_Signal_Zero_t vZeros[maxZeros];
    SmartField_Signal_Zero_t iZeros[maxZeros];
    uint32_t nZerosVoltage=findZeros(SmartField_Voltage, vZeros, maxZeros);
    uint32_t nZerosCurrent=findZeros(SmartField_Current, iZeros, maxZeros);

    uint32_t zerosFound = nZerosVoltage < nZerosCurrent ? nZerosVoltage : nZerosCurrent;

    SmartField_Result.angle = timeDomainPhase(vZeros, iZeros, zerosFound);

    //Calculate complex impedance
    float Z = (SmartField_Result.VRMS / SmartField_Result.IRMS);
	SmartField_Result.ZC = Z * exp(I * 2*M_PI * SmartField_Result.angle/360);
}

/**
 * @brief Calculate impedance using DFT
 *
 * @param	settings	SmartField configuration pointer
 *
 * Checks data using checkData() and calculates DFT using goertzel().
 */
void SmartField_Core_ProcessData_DFT(SmartField_Settings_t *settings) {
	checkData();
	/* Calculate frequency bin */
	float samplePerPeriod = 1.0 / settings->normalisedFrequency;
	uint32_t k = SMARTFIELD_ADC_BUFFER_LENGTH/samplePerPeriod;
	float _Complex Uc = goertzel(SmartField_ADC_RAW_Buffer_Byte, SMARTFIELD_VOLTAGE_COL, k) * SMARTFIELD_Q;
	float _Complex Ic = goertzel(SmartField_ADC_RAW_Buffer_Byte, SMARTFIELD_CURRENT_COL, k) * SMARTFIELD_Q / settings->Zsns;
	SmartField_Result.VRMS = (cabsf(Uc) * 2 / SMARTFIELD_ADC_BUFFER_LENGTH);
	SmartField_Result.IRMS = (cabsf(Ic) * 2 / SMARTFIELD_ADC_BUFFER_LENGTH);
	SmartField_Result.ZC = Uc / -Ic;
	SmartField_Result.angle = cargf(SmartField_Result.ZC) * 180 / M_PI;
}

/**
 * @brief Get filtered data from buffer
 *
 * @param	callbackPtr	Function pointer to be called.
 *
 * Function calls callback given in pointer, providing data buffer location and length.
 */
void SmartField_Core_DumpData(void (*callbackPtr)(float* voltage, float* current, uint32_t dataLength)){
	if(callbackPtr!=NULL){
		callbackPtr(SmartField_Voltage, SmartField_Current, SMARTFIELD_FILT_BUFFER_LENGTH);
	}
}

/**
 * @brief Get last measurement result
 *
 * @return Pointer to result buffer containing last measurement
 */
SmartField_Result_t* SmartField_Core_GetResult(void){
	return &SmartField_Result;
}

/* Private function definitions ----------------------------------------------*/

/**
 * @brief Check sampled data for errors
 *
 * Checks both voltage and current vectors using checkVector()
 */
static SmartField_Status_t checkData(void){
	SmartField_Status_t stsVoltage=SMARTFIELD_OK;
	SmartField_Status_t stsCurrent=SMARTFIELD_OK;

	SmartField_printf_Core("Checking voltage data...\r\n");
	stsVoltage = checkVector(SmartField_ADC_RAW_Buffer_Byte, SMARTFIELD_VOLTAGE_COL);
	#ifdef DEBUG
		switch (stsVoltage) {
		case SMARTFIELD_OK:
			SmartField_printf_Core("%sData OK. %s",terminalColors.green,terminalColors.reset);
			break;
		case SMARTFIELD_ADC_MALFUNCTION:
			SmartField_printf_Core("\033[0;31mToo many same samples (DC section). ADC malfunction.\033[0m\r\n");
			break;
		case SMARTFIELD_ADC_SATURATION:
			SmartField_printf_Core("\033[0;31mTwo consecutive maximum samples. Possible ADC saturation.\033[0m\r\n");
			break;
		default:
			SmartField_printf_Core("\033[0;31mUnknown data error.\r\n");
		}
	#endif
//	hSmartField.ADC_Voltage_Status = stsVoltage;
	uint8_t max = 0;
	uint8_t min = 255;
	for (uint32_t i = 0; i < SMARTFIELD_ADC_BUFFER_LENGTH; i++) {
		max = SmartField_ADC_RAW_Buffer[i].voltage > max ? SmartField_ADC_RAW_Buffer[i].voltage : max;
		min = SmartField_ADC_RAW_Buffer[i].voltage < min ? SmartField_ADC_RAW_Buffer[i].voltage : min;
	}
	printf("Voltage max=%d, min=%d\r\n", max, min);

	SmartField_printf_Core("Checking current data...\r\n");
	stsCurrent = checkVector(SmartField_ADC_RAW_Buffer_Byte, SMARTFIELD_CURRENT_COL);
	#ifdef DEBUG
		switch (stsCurrent) {
		case SMARTFIELD_OK:
			SmartField_printf_Core("%sData OK. %s",terminalColors.green,terminalColors.reset);
			break;
		case SMARTFIELD_ADC_MALFUNCTION:
			SmartField_printf_Core("Too many same samples (DC section). ADC malfunction.\r\n");
			break;
		case SMARTFIELD_ADC_SATURATION:
			SmartField_printf_Core("Two consecutive maximum samples. Possible ADC saturation.\r\n");
			break;
		default:
			SmartField_printf_Core("Unknown data error.\r\n");
		}
	#endif
//	hSmartField.ADC_Current_Status = stsCurrent;
	max = 0;
	min = 255;
	for (uint32_t i = 0; i < SMARTFIELD_ADC_BUFFER_LENGTH; i++) {
		max = SmartField_ADC_RAW_Buffer[i].current > max ? SmartField_ADC_RAW_Buffer[i].current : max;
		min = SmartField_ADC_RAW_Buffer[i].current < min ? SmartField_ADC_RAW_Buffer[i].current : min;
	}
	printf("Current max=%d, min=%d\r\n", max, min);

	return (stsVoltage==SMARTFIELD_OK && stsCurrent==SMARTFIELD_OK) ? SMARTFIELD_OK : SMARTFIELD_ERR;
}

/**
 * @brief Verifies measured data - single vector
 *
 * @param pArray	Pointer to sample data buffer
 * @param colIndex	Voltage/Current index, member of \ref SmartField_Buffer_Column_t
 *
 * Checks for long DC sections, underflow and overflow.
 * Data are read from library static buffer.
 */
static SmartField_Status_t checkVector(uint8_t pArray[][2],  uint8_t colIndex){
	uint16_t sameCnt = 0;
	for (uint32_t i = 0; (i < SMARTFIELD_ADC_BUFFER_LENGTH && sameCnt < 1000) ; i++) {
		if (pArray[i][colIndex] == pArray[i+1][colIndex]) {
			sameCnt++;
		}else{
			sameCnt=0; // Restart counter in case of value step.
		}
		if (pArray[i][colIndex]==UINT8_MAX && pArray[i+1][colIndex]==UINT8_MAX && i!=(SMARTFIELD_ADC_BUFFER_LENGTH-1) ){
			return SMARTFIELD_ADC_SATURATION;
		}
		if (pArray[i][colIndex]==0 && pArray[i+1][colIndex]==0 && i!=(SMARTFIELD_ADC_BUFFER_LENGTH-1) ){
			return SMARTFIELD_ADC_SATURATION;
		}
	}

	return sameCnt < 1000 ? SMARTFIELD_OK : SMARTFIELD_ADC_MALFUNCTION;
}



/**
 * @brief User API function - Filters measured data
 *
 * Uses FIR filter to remove DC offset and higher harmonics which
 * results in higher measurement stability and overcomes effects of saturation
 * or oscillatory problems. These should not occur, but might...
 *
 * Data are read from and saved to library static buffer.
 */
static void filterData(uint8_t pData[][2], uint8_t colIndex, float *pResult){
//	printf("%3d%%",0);fflush(stdout);
	for (int32_t n = SMARTFIELD_FIR_LEN; n < (SMARTFIELD_FIR_LEN+SMARTFIELD_FILT_BUFFER_LENGTH); n++) {
		float acc=0;
		for (int16_t k = 0; k < SMARTFIELD_FIR_LEN; k++) {
			acc += firCoef[k] * pData[n - k][colIndex];
		}
		pResult[n-SMARTFIELD_FIR_LEN]=acc;
//		#ifdef DEBUG
//			if(n%102==0){
//				printf("\b\b\b\b%3ld%%", (100*(n-SMARTFIELD_FIR_LEN))/SMARTFIELD_FILT_BUFFER_LENGTH );fflush(stdout);
//			}
//		#endif
	}
//	#ifdef DEBUG
//		printf("\b\b\b\b%3d%%\r\n",100);
//	#endif
}

static float timeDomainPhase(SmartField_Signal_Zero_t *x, SmartField_Signal_Zero_t*y, uint32_t zerosFound){

    /* Exit if less then one full period was found */
    if(zerosFound<3){
        return NAN;
    }

    /* Calculate mean period */
    float period=0;
    for (uint32_t i=0 ; i < (zerosFound-1)/2 ; i++){
        period += x[2*i+2].interpPos - x[2*i].interpPos;
    }
    period = period / ((zerosFound-1)/2);

    /* Find first complete phase information */
    if(x[0].rising != y[0].rising){
        x += 1;
        zerosFound -= 1;
    }

    /* Evaluate mean delay */
    float delay=0;
    for (uint32_t i=0 ; i < (zerosFound-1)/2 ; i++){
        delay += (y[2*i].interpPos - x[2*i].interpPos + y[2*i+1].interpPos - x[2*i+1].interpPos)/2;
    }
    delay = delay / ((zerosFound-1)/2);

    return 360 * delay / period;
}

static uint32_t findZeros(float *vector, SmartField_Signal_Zero_t *zeros, uint32_t maxZeros){
    uint32_t zerosFound=0;
    memset(zeros,0,maxZeros*sizeof(SmartField_Signal_Zero_t));
    for (uint16_t i = 0; i < SMARTFIELD_FILT_BUFFER_LENGTH && zerosFound < maxZeros; i++) {
        if( signbit(vector[i]) != signbit(vector[i+1]) ){
             zeros[zerosFound].rising = signbit(vector[i])==0 ? false : true;
             float dy = (vector[i + 1] - vector[i]);
             zeros[zerosFound++].interpPos = i - vector[i] / dy;
        }
    }
    return zerosFound;
}

/**
 * @brief Calculates single bin DFT using Goertzel filter
 *
 */
static float _Complex goertzel(uint8_t pData[][2], uint8_t colIndex, uint32_t k) {
	uint32_t N = SMARTFIELD_ADC_BUFFER_LENGTH;
	float Ak = 2 * cos(2*M_PI*k/N);
	float _Complex WN = exp(-I*2*M_PI*k/N);
	float s0 = 0, s1 = 0, s2 = 0;
    for (uint32_t i = 0; i < N; i++) {
        s2 = s1;
        s1 = s0;
		s0 = pData[i][colIndex] + Ak * s1 - s2;		
	}
	return s0 - WN * s1;
}







